import torch
import numpy as np
import pyro
import sys
from tqdm import tqdm

class Model_LVM():
    def __init__(self, device="cpu", dtype="float32"):
        device, dtype = self._device_and_dtype(device, dtype)
        self.device = device
        self.dtype = dtype
        self.pyro = pyro
        if self.device.type == 'cuda' and torch.cuda.is_available():
            torch.set_default_tensor_type('torch.cuda.FloatTensor')
        else:
            torch.set_default_tensor_type('torch.FloatTensor')
    
    @staticmethod
    def _device_and_dtype(device, dtype):
        """define device and dtype

        parse device and dtype into torch.dtype and torch.device

        # Arguments
            device: str or torch.device, "cpu" or "gpu"
            dtype: str or torch.dtype, "float32" and "float64" are supported

        # Example

        ```python
            device, dtype = _device_and_dtype(torch.device("cpu"), "float32")
            device, dtype = _device_and_dtype("gpu", "float32")
        ````

        """
        if type(device) is int:
            device = torch.device('cuda:'+str(device))

        if type(device) is str:
            device = torch.device(device)

        if type(dtype) is not str:
            return device, dtype
        else:
            if dtype == "float32":
                return device, torch.float32
            if dtype == "float64":
                return device, torch.float64
        return device, torch.float32
    
    def _get_DataLoader(self, X, Y=None,SP=None,RE=None, batch_size=25, shuffle=True, parallel=0, drop_last=True):
        if self.device.type == 'cuda':
            torch.cuda.set_device(self.device)
            pin_memory = False
        else:
            pin_memory = True
        #init_func = lambda: torch.multiprocessing.set_start_method('spawn', True)
        if type(RE) is np.ndarray:
            if type(Y) is np.ndarray:
                if type(SP) is np.ndarray:
                    data = torch.utils.data.TensorDataset(torch.tensor(X.copy(), dtype=torch.float32, device=torch.device('cpu')), 
                                                          torch.tensor(Y.copy(), dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(SP, dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.asarray(RE).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')),
                                                          torch.tensor(np.arange(0, X.shape[0]).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))
                else:
                    data = torch.utils.data.TensorDataset(torch.tensor(X.copy(), dtype=torch.float32, device=torch.device('cpu')), 
                                                          torch.tensor(Y.copy(), dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.asarray(RE).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')),
                                                          torch.tensor(np.arange(0, X.shape[0]).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))

            else:
                if type(SP) is np.ndarray:
                    data = torch.utils.data.TensorDataset(torch.tensor(X.copy(), dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(SP, dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.asarray(RE).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')),
                                                          torch.tensor(np.arange(0, X.shape[0]).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))
                else: 
                    data = torch.utils.data.TensorDataset(torch.tensor(X.copy(), dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.asarray(RE).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')),
                                                          torch.tensor(np.arange(0, X.shape[0]).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))
        
        else: 
            if type(Y) is np.ndarray:
                if type(SP) is np.ndarray:
                    data = torch.utils.data.TensorDataset(torch.tensor(X.copy(), dtype=torch.float32, device=torch.device('cpu')), 
                                                        torch.tensor(Y.copy(), dtype=torch.float32, device=torch.device('cpu')),
                                                        torch.tensor(SP, dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.arange(0, X.shape[0]).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))
                else:
                    data = torch.utils.data.TensorDataset(torch.tensor(X.copy(), dtype=torch.float32, device=torch.device('cpu')), 
                                                        torch.tensor(Y.copy(), dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.arange(0, X.shape[0]).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))

            else:
                if type(SP) is np.ndarray:
                    data = torch.utils.data.TensorDataset(torch.tensor(X.copy(), dtype=torch.float32, device=torch.device('cpu')),
                                                        torch.tensor(SP, dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.arange(0, X.shape[0]).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))
                else: 
                    data = torch.utils.data.TensorDataset(torch.tensor(X.copy(), dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.arange(0, X.shape[0]).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))          

        DataLoader = torch.utils.data.DataLoader(data, batch_size=batch_size, shuffle=shuffle, num_workers=int(parallel), pin_memory=pin_memory, drop_last=drop_last)
        torch.cuda.empty_cache()
        return DataLoader
    
    def build(self, input_shape, out, scale_mu = 5.0, scale_lf = 1.0,scale_lv=1.0,  df=2,
              guide='LaplaceApproximation', 
              family = "binomial", link = "logit"):

        self.create_link(family, link)
            
        #self.link = link
        torch.cuda.empty_cache()
        pyro.enable_validation(True)
        pyro.clear_param_store()
        #XX = torch.tensor(X, dtype = torch.float32)
        #YY = torch.tensor(Y, dtype = torch.float32)
        self.e = input_shape[1]
        self.sp = out
        self.n = input_shape[0]
        self.df = df
        e = self.e
        sp = self.sp
        n = self.n

        def model(XX, YY=None, indices=None, posterior = self.posterior):
            mu_scale = torch.ones([e, sp])*scale_mu
            mu_loc = torch.zeros([e, sp])
            lv_scale2 = torch.ones([n, df])*scale_lv
            lv_loc = torch.zeros([n, df])
            lf_scale2 = torch.ones([df, sp])*scale_lf
            lf_loc = torch.zeros([df, sp])
            mu = pyro.sample("mu", pyro.distributions.Normal(mu_loc, mu_scale).to_event())
            lf = pyro.sample("lf", pyro.distributions.Normal(lf_loc, lf_scale2).to_event())
            lv = pyro.sample("lv", pyro.distributions.Normal(lv_loc, lv_scale2).to_event())
            with pyro.plate('data', size = XX.shape[0]):
                loc_tmp = XX.matmul(mu).add( lv.index_select(0, indices).matmul(lf) )
                posterior(loc_tmp, YY)
        self.model = model

        guide2 = guide
        self.elbo = pyro.infer.JitTrace_ELBO(ignore_jit_warnings=True)
        if guide2 == 'LaplaceApproximation':
            self.guide = pyro.infer.autoguide.AutoLaplaceApproximation(self.model)

        if guide2 == 'LowRankMultivariateNormal':
            self.guide = pyro.infer.autoguide.AutoLowRankMultivariateNormal(self.model, init_loc_fn=pyro.infer.autoguide.init_to_mean)
            self.elbo = pyro.infer.Trace_ELBO(ignore_jit_warnings=True)

        if guide2 == "Delta":
            self.guide = pyro.infer.autoguide.AutoDelta(self.model)

        if guide2 == 'DiagonalNormal':
            self.guide = pyro.infer.autoguide.AutoDiagonalNormal(self.model)
    


    def create_link(self, family="binomial", link="logit"):
        if family == "binomial":
            if link == "logit":
                self.link = lambda value: torch.sigmoid(value)
                self.posterior = lambda loc, YY: pyro.sample("obs", pyro.distributions.Binomial(1, probs = torch.sigmoid(loc)).to_event(), obs = YY)
                self.logLik = lambda loc, YY: pyro.distributions.Binomial(1, probs = torch.sigmoid(loc)).log_prob(YY)
            if link == "probit":
                normal = torch.distributions.Normal(0.0, 1.0)
                self.link = lambda value: normal.cdf(value)
                self.posterior = lambda loc, YY: pyro.sample("obs", pyro.distributions.Binomial(1, probs = normal.cdf(loc)).to_event(), obs = YY)
                self.logLik = lambda loc, YY:  pyro.distributions.Binomial(1, probs = normal.cdf(loc)).log_prob(YY)
        if family == "poisson":
            if link == "log":
                self.link = lambda value: value.exp()
                self.posterior = lambda loc, YY: pyro.sample("obs", pyro.distributions.Poisson(rate= loc.exp()).to_event(), obs = YY)
                self.logLik = lambda loc, YY: pyro.distributions.Poisson(rate= loc.exp()).log_prob(YY)
            if link == "linear":
                self.link = lambda value: torch.clamp(value, min=0.00000001)
                self.posterior = lambda loc, YY: pyro.sample("obs", pyro.distributions.Poisson(rate= torch.clamp(loc, min=0.00000001)).to_event(), obs = YY)
                self.logLik = lambda loc, YY: pyro.distributions.Poisson(rate= torch.clamp(loc, min=0.00000001)).log_prob(YY)

    def fit(self, X, Y,
            lr = [0.1], epochs = 200,
            batch_size = 25,
            num_samples=100,
            parallel = 0):
        
        if self.device.type == 'cuda' and torch.cuda.is_available():
            torch.set_default_tensor_type('torch.cuda.FloatTensor')
        else:
            torch.set_default_tensor_type('torch.FloatTensor')

 
        
        if len(lr) == 1:
             adam = pyro.optim.Adam({'lr' : lr[0]})
        else:
            def per_param_callable(module_name, param_name):
                if param_name == 'mu':
                    return {"lr": lr[0]}
                if param_name == 'lf':
                    return {"lr": lr[1]}
                else:
                    return {"lr": lr[2]}
            adam = pyro.optim.Adam(per_param_callable)

        
        self.svi = pyro.infer.SVI(self.model, self.guide, adam, loss = self.elbo)
        stepSize = np.floor(X.shape[0] / batch_size).astype(int)
        dataLoader = self._get_DataLoader(X, Y, batch_size = batch_size, shuffle=True, parallel=parallel)
        batch_loss = np.zeros(stepSize)
        self.history = np.zeros(epochs)
        desc='loss: Inf'
        ep_bar = tqdm(range(epochs),bar_format= "Iter: {n_fmt}/{total_fmt} {l_bar}{bar}| [{elapsed}, {rate_fmt}{postfix}]", file=sys.stdout)
        for epoch in ep_bar:
            for step, (x, y, ind) in enumerate(dataLoader):
                x = x.to(self.device, non_blocking=True)
                y = y.to(self.device, non_blocking=True)
                ind = ind.to(self.device, non_blocking=True).view([-1])
                loss = self.svi.step(x, y, ind)
                batch_loss[step] = loss
            bl = np.mean(batch_loss)
            bl = np.round(bl, 3)
            ep_bar.set_postfix(loss=f'{bl}')
        self.posterior_samples = pyro.infer.Predictive(self.model, guide=self.guide, num_samples=num_samples)(torch.tensor(X.copy(), dtype=torch.float32), 
                                                                                                              torch.tensor(Y.copy(), dtype=torch.float32),
                                                                                                              torch.tensor(np.arange(0, X.shape[0]), dtype=torch.long))

        
        self.lf = self.posterior_samples["lf"].data.squeeze(dim=1).mean(dim=0)
        self.lv = self.posterior_samples["lv"].data.squeeze(dim=1).mean(dim=0)
        self.mu = self.posterior_samples["mu"].data.squeeze(dim=1).mean(dim=0)

    def getLogLik(self, X, Y, batch_size = 25, parallel=0):
        dataLoader = self._get_DataLoader(X = X, Y = Y, batch_size = batch_size, shuffle = False, parallel = parallel, drop_last = False)
        ll = 0
        for step, (x, y, ind) in enumerate(dataLoader):
            x = x.to(self.device, non_blocking=True)
            y = y.to(self.device, non_blocking=True)
            ind = ind.to(self.device, non_blocking=True).view([-1])
            lin = self.link(x.matmul(self.mu).add( self.lv.index_select(0, ind).matmul(self.lf) ))
            ll += self.logLik(lin, y).sum().data.cpu().numpy()
        return ll
    
    def predict(self, newdata, batch_size = 25, parallel=0, mean_field=True ):
        dataLoader = self._get_DataLoader(X = newdata, Y = None, batch_size = batch_size, shuffle = False, parallel = parallel, drop_last = False)
        pred = []

        if mean_field:
            for step, (x, ind) in enumerate(dataLoader):
                x = x.to(self.device, non_blocking=True)
                lin = self.link(x.matmul(self.mu))
                pred.append(lin)
            return torch.cat(pred, dim=0).data.cpu().numpy()
        else:
            mu = self.posterior_samples["mu"].squeeze(dim=1)
            for step, (x, ind) in enumerate(dataLoader):
                x = x.to(self.device, non_blocking=True)
                lin = self.link( x.expand([mu.shape[0],batch_size,newdata.shape[1]]).matmul(mu) )
                pred.append(lin)
            return torch.cat(pred,dim=1).data.cpu().numpy()    

    def predictPosterior(self, X, batch_size = 25, parallel=0, num_samples=10):
        dataLoader = self._get_DataLoader(X = X, batch_size = batch_size, shuffle = False, parallel = parallel, drop_last = False)
        pred = []
        mu = self.posterior_samples["mu"].squeeze(dim=1)
        lf = self.posterior_samples["lf"].squeeze(dim=1)
        lv = self.posterior_samples["lv"].squeeze(dim=1)
        for step, (x, ind) in enumerate(dataLoader):
            x = x.to(self.device, non_blocking=True)
            ind = ind.to(self.device, non_blocking=True).view([-1])
            lin = self.link( x.expand([mu.shape[0],batch_size,X.shape[1]]).matmul(mu).add( lv.index_select(1, ind).matmul(lf) ) )
            pred.append(lin)
        return torch.cat(pred,dim=1).data.cpu().numpy()

    def se(self, X, Y, batch_size=25, parallel=0):
        dataLoader = self._get_DataLoader(X = X, Y = Y, batch_size = batch_size, shuffle = False, parallel = parallel, drop_last = False)
        se = []
        lf = self.lf
        lv = self.lv
        weights_base = self.mu.cpu().data.numpy()
        y_dim = Y.shape[1]

        if self.device.type == 'cuda':
            device = self.device.type+ ":" + str(self.device.index)
        else:
            device = 'cpu'
        
        _ = sys.stdout.write("\nCalculating standard errors...\n")

        desc='loss: Inf'
        sp_bar = tqdm(range(Y.shape[1]),bar_format= "Species: {n_fmt}/{total_fmt} {l_bar}{bar}| [{elapsed}, {rate_fmt}]", file=sys.stdout)

        for i in sp_bar:
            #_ = sys.stdout.write("\rSpecies: {}/{} ".format(i+1, y_dim))
            sys.stdout.flush()
            weights = torch.tensor(weights_base[:,i].reshape([-1,1]), device=self.device, dtype=self.dtype, requires_grad=True).to(self.device)
            if i == 0:
                constants = torch.tensor(weights_base[:,(i+1):], device=self.device, dtype=self.dtype).to(self.device)
                w = torch.cat([weights, constants], dim=1)
            elif i < y_dim:
                w = torch.cat([torch.tensor(weights_base[:,0:i], device=self.device, dtype=self.dtype).to(self.device), 
                               weights, 
                               torch.tensor(weights_base[:,(i+1):], device=self.device, dtype=self.dtype).to(self.device)],dim=1)
            else:
                constants = torch.tensor(weights_base[:,0:i], device=self.device, dtype=self.dtype).to(self.device)
                w = torch.cat([constants, weights], dim=1)
            for step, (x, y, ind) in enumerate(dataLoader):
                x = x.to(self.device, non_blocking=True)
                y = y.to(self.device, non_blocking=True)
                ind = ind.to(self.device, non_blocking=True).view([-1])

                pred = x.matmul(w).add(lv.index_select(0, ind).matmul(lf))
                loss = self.logLik(self.link(pred), y).sum().neg()

                first_gradients = torch.autograd.grad(loss, weights, retain_graph=True, create_graph=True, allow_unused=True)
                second = []
                for j in range(x.shape[1]):
                    second.append(torch.autograd.grad(first_gradients[0][j,0], inputs = weights, retain_graph = True, create_graph = False, allow_unused = False)[0])
                    hessian = torch.cat(second,dim=1)

                if step < 1:
                    hessian_out = hessian
                else:
                    hessian_out += hessian
            se.append(torch.sqrt(torch.diag(torch.inverse(hessian_out))).data.cpu().numpy())
        return se

    @property
    def covariance(self):
        return self.lf.t().matmul(self.lf).data.cpu().numpy()

    @property
    def weights(self):
        return self.mu.data.cpu().numpy()

    @property
    def lfs(self):
        return self.lf.data.cpu().numpy()

    @property
    def lvs(self):
        return self.lv.data.cpu().numpy()

    def set_posterior_samples(self, ps):
        self.posterior_samples = ps


import numpy as np
import torch
import itertools
import sys
from typing import Union, Tuple, List, Optional, Callable
from tqdm import tqdm
from torch import nn, optim

class Model_sjSDM:
    def __init__(self, device: str = "cpu", dtype: str = "float32"):
        """sjSDM constructor

        Args:
            device (str, optional): Device type. Defaults to "cpu".
            dtype (str, optional): Dtype. Defaults to "float32".

        """
        self.params = []
        self.losses = []
        self.env = None
        self.spatial = None
        device, dtype = self._device_and_dtype(device, dtype)
        self.device = device
        self.dtype = dtype
        torch.set_default_dtype(self.dtype)
        torch.set_default_tensor_type('torch.FloatTensor')

        @torch.jit.script
        def l1_loss(tensor: torch.Tensor, l1: float):
            return tensor.abs().sum().mul(l1)

        @torch.jit.script
        def l2_loss(tensor: torch.Tensor, l2: float):
            return tensor.pow(2.0).sum().mul(l2)
        
        self.l1_l2 = [l1_loss, l2_loss]

    def __call__(self):
        pass

    
    def __repr__(self):
        return "Model_base: \n  \n"

    @staticmethod
    def _device_and_dtype(device: Union[int, str, torch.device], dtype: Union[str, torch.dtype]) -> Tuple[torch.device, torch.dtype]:
        """Device and dtype parster

        Args:
            device (Union[int, str, torch.device]): device
            dtype (Union[str, torch.dtype]): dtype

        Returns:
            Tuple[torch.device, torch.dtype]: Tuple of device and dtype
        """        

        if type(device) is int:
            device = torch.device('cuda:'+str(device))

        if type(device) is str:
            device = torch.device(device)

        if type(dtype) is not str:
            return device, dtype

        if dtype == "float32":
            return device, torch.float32
        if dtype == "float64":
            return device, torch.float64
        return device, torch.float32


    def add_env(self, 
                input_shape: int, 
                output_shape: int, 
                hidden: List = [], 
                activation: List[str] =['linear'], 
                bias: List[bool] = [False], 
                l1: float = -99, 
                l2: float = -99, 
                dropout: float = -99) -> None:
        """Add environmental model

        Args:
            input_shape (int): number of predictors
            output_shape (int): number of species
            hidden (List, optional): List of units for hidden layers (each entry corresponds to one hidden layer). Defaults to [].
            activation (List[str], optional): List of activation functions. Defaults to ['linear'].
            bias (List[bool], optional): Bias in each hidden layer. Defaults to [False].
            l1 (float, optional): LASSO regularization. Defaults to -99.
            l2 (float, optional): Ridge regularization. Defaults to -99.
            dropout (float, optional): [Dropoutrate. Defaults to -99.
        """                
        
        self.env = (self._build_NN(input_shape, output_shape, hidden,bias, activation, dropout))
        self.params.append(self.env.parameters())
        self.input_shape = input_shape
        self.output_shape = output_shape
        for p in self.env.parameters():
            if l1 > 0.0: 
                self.losses.append(lambda: self.l1_l2[0](p, l1))
            if l2 > 0.0:
                self.losses.append(lambda: self.l1_l2[1](p, l2))

    def add_spatial(self, 
                    input_shape: int, 
                    output_shape: int, 
                    hidden: List = [], 
                    activation: List[str] = ['linear'], 
                    bias: List[bool] = [False], 
                    l1: float = -99, 
                    l2: float = -99, 
                    dropout: float = -99) -> None:
        """Add spatial model

        Args:
            input_shape (int): number of predictors
            output_shape (int): number of species
            hidden (List, optional): List of units for hidden layers (each entry corresponds to one hidden layer). Defaults to [].
            activation (List[str], optional): List of activation functions. Defaults to ['linear'].
            bias (List[bool], optional): Bias in each hidden layer. Defaults to [False].
            l1 (float, optional): LASSO regularization. Defaults to -99.
            l2 (float, optional): Ridge regularization. Defaults to -99.
            dropout (float, optional): [Dropoutrate. Defaults to -99.
        """                    
        self.spatial = (self._build_NN(input_shape, output_shape, hidden, bias, activation, dropout))
        self.params.append(self.spatial.parameters())
        for p in self.spatial.parameters():
            if l1 > 0.0: 
                self.losses.append(lambda: self.l1_l2[0](p, l1))
            if l2 > 0.0:
                self.losses.append(lambda: self.l1_l2[1](p, l2))
                
    def _build_NN(self, 
                  input_shape: int, 
                  output_shape: int, 
                  hidden: List, 
                  bias: List[bool], 
                  activation: List[str], 
                  dropout: float) -> torch.nn.modules.container.Sequential:
        """Build neural network

        Args:
            input_shape (int): Number of predictors
            output_shape (int): Number of species
            hidden (List): List of hidden layers
            bias (List[bool]): Biases in hidden layers
            activation (List[str]): List of activation functions
            dropout (float): Dropout rate

        Returns:
            torch.nn.modules.container.Sequential: Sequential neural network object
        """                  
        model_list = nn.ModuleList()
        if len(hidden) != len(activation):
            activation = [activation[0] for _ in range(len(hidden))]

        if len(bias) == 1:
            bias = [bias[0] for _ in range(len(hidden)+1)]  

        if len(hidden) > 0:
            for i in range(len(hidden)):
                if i == 0:
                    model_list.append(nn.Linear(input_shape, hidden[i], bias=bias[i]).type(self.dtype))
                else:
                    model_list.append(nn.Linear(hidden[i-1], hidden[i], bias=bias[i]).type(self.dtype))

                if activation[i] == "relu":
                     model_list.append(nn.ReLU())
                if activation[i] == "selu":
                     model_list.append(nn.SELU())
                if activation[i] == "leakyrelu":
                     model_list.append(nn.LeakyReLU())
                if activation[i] == "tanh": 
                    model_list.append(nn.Tanh())
                if activation[i] == "sigmoid":
                    model_list.append(nn.Sigmoid())
                if dropout > 0.0:
                    model_list.append(nn.Dropout(p=dropout))

        if len(hidden) > 0:
            model_list.append(nn.Linear(hidden[-1], output_shape, bias=bias[-1]).type(self.dtype))
        else:
            model_list.append(nn.Linear(input_shape, output_shape, bias=False).type(self.dtype))
        return nn.Sequential(*model_list)
    
    def _get_DataLoader(self, 
                        X: np.ndarray, 
                        Y: Optional[np.ndarray] = None,
                        SP: Optional[np.ndarray] = None,
                        batch_size: int = 25, 
                        shuffle: bool = True, 
                        parallel: int = 0, 
                        drop_last: bool = True) -> torch.utils.data.DataLoader:
        """Create dataloader

        Args:
            X (np.ndarray): Environment (n*p)
            Y (Optional[np.ndarray], optional): Species occurence matrix (n*s). Defaults to None.
            SP (Optional[np.ndarray], optional): Spatial matrix (n*sp). Defaults to None.
            batch_size (int, optional): Batch size. Defaults to 25.
            shuffle (bool, optional): Shuffle data or not. Defaults to True.
            parallel (int, optional): Parallelize data fetching or not. Defaults to 0.
            drop_last (bool, optional): Drop last rows in left-over batch or not. Defaults to True.

        Returns:
            torch.utils.data.DataLoader: DataLoader object
        """                        
        # reticulate creates non writeable arrays
        X = X.copy()
        if type(Y) is np.ndarray:
            Y = Y.copy()
        if type(SP) is np.ndarray:
            SP = SP.copy()

        if self.device.type == 'cuda':
            torch.cuda.set_device(self.device)
            pin_memory = False
        else:
            pin_memory = True

        if type(Y) is np.ndarray:
            if type(SP) is np.ndarray:
                data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=self.dtype, device=torch.device('cpu')), 
                                                      torch.tensor(Y, dtype=self.dtype, device=torch.device('cpu')),
                                                      torch.tensor(SP, dtype=self.dtype, device=torch.device('cpu')))
            else:
                data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=self.dtype, device=torch.device('cpu')), 
                                                    torch.tensor(Y, dtype=self.dtype, device=torch.device('cpu')))
        else:
            if type(SP) is np.ndarray:
                data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=self.dtype, device=torch.device('cpu')),
                                                    torch.tensor(SP, dtype=self.dtype, device=torch.device('cpu')))
            else: 
                data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=self.dtype, device=torch.device('cpu')))            

        DataLoader = torch.utils.data.DataLoader(data, batch_size=batch_size, shuffle=shuffle, num_workers=int(parallel), pin_memory=pin_memory, drop_last=drop_last)
        torch.cuda.empty_cache()
        return DataLoader

    def build(self, 
              df: Optional[int] = None,
              optimizer: Optional[Callable] = None, 
              l1: float = 0.0, 
              l2: float = 0.0,
              reg_on_Cov: bool = True, 
              reg_on_Diag: bool = True, 
              inverse: bool = False, 
              link: str = "probit",
              alpha: float = 1.0, 
              diag: bool = False, 
              scheduler: bool = True,
              patience: int = 2, 
              factor: float = 0.95,
              mixed: bool = False) -> None:
        """Build and specify sjSDM Model

        Args:
            df (Optional[int], optional): Sigma is parametrized by a [sp, df] matrix. Defaults to None.
            optimizer (Optional[Callable], optional): Optimizer used for stochastic gradient descent. Defaults to None.
            l1 (float, optional): LASSO regularization on sigma. Defaults to 0.0.
            l2 (float, optional): Ridge regularization on sigma. Defaults to 0.0.
            reg_on_Cov (bool, optional): Regularization on sigma or on it's square-root matrix (used to parametrize sigma). Defaults to True.
            reg_on_Diag (bool, optional): Regularization on diagonals or not. Defaults to True.
            inverse (bool, optional): Regularization on the precision matrix or not. Defaults to False.
            link (str, optional): which link/family to use. Defaults to "probit".
            alpha (float, optional): 1.0 corresponds to logit and 1.70169 to probit. Defaults to 1.0.
            diag (bool, optional): Use only diagonal matrix or not. Defaults to False.
            scheduler (bool, optional): Use learning rate scheduler or not. Defaults to True.
            patience (int, optional): Number of epochs for lr scheduler. Defaults to 2.
            factor (float, optional): Rate to decrease learning rate. Defaults to 0.95.
            mixed (bool, optional): use mixed half-precision training or not. Defaults to False.

        """              
        
        if self.device.type == 'cuda' and torch.cuda.is_available():
            torch.set_default_tensor_type('torch.cuda.FloatTensor')
            self.env.cuda(self.device)
            if self.spatial is not None:
                self.spatial.cuda(self.device)
        else:
            torch.set_default_tensor_type('torch.FloatTensor')
        
        self.link = link
        self.df = df
        self.l1 = l1 #test
        self.l2 = l2 #test
        self.alpha = alpha
        self.mixed = mixed
        r_dim = self.output_shape
        if not diag:
            low = -np.sqrt(6.0/(r_dim+df))
            high = np.sqrt(6.0/(r_dim+df))               
            self.sigma = torch.tensor(np.random.uniform(low, high, [r_dim, df]), requires_grad = True, dtype = self.dtype, device = self.device).to(self.device)
            self._loss_function = self._build_loss_function()
            self._build_cov_constrain_function(l1 = l1, l2 = l2, reg_on_Cov = reg_on_Cov, reg_on_Diag = reg_on_Diag, inverse = inverse)
            self.params.append([self.sigma])
        else:
            self.sigma = torch.zeros([r_dim, r_dim], dtype = self.dtype, device = self.device).to(self.device)
            self.df = r_dim
            self._loss_function = self._build_loss_function()
            self._build_cov_constrain_function(l1 = l1, l2 = l2, reg_on_Cov = reg_on_Cov, reg_on_Diag = reg_on_Diag, inverse = inverse)

        if optimizer != None:
            self.optimizer = optimizer(params = itertools.chain(*self.params))

        if scheduler:
            self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(self.optimizer, mode="min", patience=patience, factor=factor, verbose=True)
            self.useSched = True
        else:
            self.useSched = False

    def fit(self, 
            X: np.ndarray, 
            Y: np.ndarray, 
            SP: Optional[np.ndarray] = None, 
            batch_size: int = 25, 
            epochs: int = 100, 
            sampling: int = 100, 
            parallel: int = 0,
            early_stopping_training: int = -1) -> None:
        """Fit sjSDM model

        Args:
            X (np.ndarray): Environment, n*p matrix
            Y (np.ndarray): Species occurrence matrix, n*s matrix
            SP (Optional[np.ndarray], optional): Spatial predictors, n*sp matrix. Defaults to None.
            batch_size (int, optional): Batch size for stochastic gradient descent. Defaults to 25.
            epochs (int, optional): Number of epochs. Defaults to 100.
            sampling (int, optional): Number of MC samples for each species. Defaults to 100.
            parallel (int, optional): Use parallelization for DataLoader. Defaults to 0.
            early_stopping_training (int, optional): Use early stopping or not. Defaults to -1.
        """            
        stepSize = np.floor(X.shape[0] / batch_size).astype(int)
        dataLoader = self._get_DataLoader(X, Y, SP, batch_size, True, parallel)
        any_losses = len(self.losses) > 0
        batch_loss = np.zeros(stepSize)
        self.history = np.zeros(epochs)

        df = self.df
        alpha = self.alpha
        
        if self.device.type == 'cuda':
            device = self.device.type+ ":" + str(self.device.index)
        else:
            device = 'cpu'

        early_stopping_training_loss = np.inf
        counter_early_stopping_training = 0

        if early_stopping_training > 0:
            early_stopping_training_boolean = True
        else:
            early_stopping_training_boolean = False

        mixed = self.mixed

        if mixed:
            scaler = torch.cuda.amp.GradScaler()
            def update_func(loss):
                scaler.scale(loss).backward()
                scaler.step(self.optimizer)
                scaler.update()
        else:
            def update_func(loss):
                loss.backward()
                self.optimizer.step()

        desc='loss: Inf'
        ep_bar = tqdm(range(epochs),bar_format= "Iter: {n_fmt}/{total_fmt} {l_bar}{bar}| [{elapsed}, {rate_fmt}{postfix}]", file=sys.stdout)
        if type(SP) is np.ndarray:
            for epoch in ep_bar:
                for step, (x, y, sp) in enumerate(dataLoader):
                    x = x.to(self.device, non_blocking=True)
                    y = y.to(self.device, non_blocking=True)
                    sp = sp.to(self.device, non_blocking=True)
                    self.optimizer.zero_grad()
                    with torch.cuda.amp.autocast(enabled=mixed):
                        mu = self.env(x) + self.spatial(sp)
                        #tmp(mu: torch.Tensor, Ys: torch.Tensor, sigma: torch.Tensor, batch_size: int, sampling: int, df: int, alpha: float
                        loss = self._loss_function(mu, y, self.sigma, batch_size, sampling, df, alpha, device, self.dtype)
                        loss = loss.mean()
                        if any_losses:
                            for k in range(len(self.losses)):
                                loss+= self.losses[k]()
                    update_func(loss)
                    batch_loss[step] = loss.item()
                bl = np.mean(batch_loss)
                bl = np.round(bl, 3)
                ep_bar.set_postfix(loss=f'{bl}')

                self.history[epoch] = bl
                if self.useSched:
                    self.scheduler.step(bl)
                
                if early_stopping_training_boolean:
                    if bl < early_stopping_training_loss:
                        early_stopping_training_loss = bl
                        counter_early_stopping_training = 0
                    else:
                        counter_early_stopping_training+=1
                    if counter_early_stopping_training == early_stopping_training:
                        _ = sys.stdout.write("\nEarly stopping...")
                        break
            self.spatial.eval()             
        else:
            for epoch in ep_bar:  
                for step, (x, y) in enumerate(dataLoader):
                    x = x.to(self.device, non_blocking=True)
                    y = y.to(self.device, non_blocking=True)
                    self.optimizer.zero_grad()
                    with torch.cuda.amp.autocast(enabled=mixed):
                        mu = self.env(x)
                        loss = self._loss_function(mu, y, self.sigma, batch_size, sampling, df, alpha, device, self.dtype)
                        loss = loss.mean()
                        if any_losses:
                            for k in range(len(self.losses)):
                                loss += self.losses[k]()
                    update_func(loss)
                    batch_loss[step] = loss.item()
                bl = np.mean(batch_loss)
                bl = np.round(bl, 3)
                ep_bar.set_postfix(loss=f'{bl}')
                self.history[epoch] = bl
                if self.useSched:
                    self.scheduler.step(bl)
                
                if early_stopping_training_boolean:
                    if bl < early_stopping_training_loss:
                        early_stopping_training_loss = bl
                        counter_early_stopping_training = 0
                    else:
                        counter_early_stopping_training+=1
                    if counter_early_stopping_training == early_stopping_training:
                        _ = sys.stdout.write("\nEarly stopping...")
                        break
        torch.cuda.empty_cache()
        self.env.eval()

        
    def logLik(self,
               X: np.ndarray,  
               Y: np.ndarray,
               SP: Optional[np.ndarray] = None, 
               batch_size: int = 25, 
               parallel: int= 0, 
               sampling: int = 100,
               individual: bool = False,
               train: bool = True) -> Tuple[float, float]:
        """Calculate log-Likelihood

        Args:
            X (np.ndarray): Environment, n*p matrix
            Y (np.ndarray): Species occurrence matrix, n*s matrix
            SP (Optional[np.ndarray], optional): Spatial predictor matrix, n*sp matrix. Defaults to None.
            batch_size (int, optional): Batch size (to prevent memory leakage). Defaults to 25.
            parallel (int, optional): Use parallelization for DataLoader. Defaults to 0.
            sampling (int, optional): Number of MC samples for each species. Defaults to 100.
            individual (bool, optional): Calculate log-LL for each species or not. Defaults to False.
            train (bool, optional): Train or evaluation mode. Defaults to True.

        Returns:
            Tuple[float, float]: Tuple of log-Likelihood and regularization loss
        """               
        dataLoader = self._get_DataLoader(X = X, Y = Y, SP=SP, batch_size = batch_size, shuffle = False, parallel = parallel, drop_last = False)
        loss_function = self._build_loss_function(train=train)
        torch.cuda.empty_cache()
        any_losses = len(self.losses) > 0

        if self.device.type == 'cuda':
            device = self.device.type+ ":" + str(self.device.index)
        else:
            device = 'cpu'

        logLik = []
        logLikReg = 0

        if type(SP) is np.ndarray:
            for step, (x, y, sp) in enumerate(dataLoader):
                x = x.to(self.device, non_blocking=True)
                y = y.to(self.device, non_blocking=True)
                sp = sp.to(self.device, non_blocking=True)
                mu = self.env(x) + self.spatial(sp)
                loss = loss_function(mu, y, self.sigma, x.shape[0], sampling, self.df, self.alpha, device, self.dtype)
                logLik.append(loss.data)
        else:
            for step, (x, y) in enumerate(dataLoader):
                x = x.to(self.device, non_blocking=True)
                y = y.to(self.device, non_blocking=True)
                mu = self.env(x)
                loss = loss_function(mu, y, self.sigma, x.shape[0], sampling, self.df, self.alpha, device, self.dtype)
                logLik.append(loss.data)

        loss_reg = torch.tensor(0.0, device=self.device, dtype=self.dtype).to(self.device)
        for k in range(len(self.losses)):
                loss_reg = loss_reg.add(self.losses[k]())

        logLikReg = loss_reg.data.cpu().numpy()
        torch.cuda.empty_cache()
        if not individual:
            logLik = torch.cat(logLik).sum().data.cpu().numpy()
        else:
            logLik = torch.cat(logLik).data.cpu().numpy()
        return logLik, logLikReg

    def predict(self, 
                newdata: Optional[np.ndarray] = None,
                SP: Optional[np.ndarray] = None,
                train: bool = False, 
                batch_size: int = 25, 
                parallel: int = 0, 
                sampling: int = 100, 
                link: bool = True) -> np.ndarray:
        """Predict with sjSDM

        Args:
            newdata (Optional[np.ndarray], optional): (New) environment, n*p matrix. Defaults to None.
            SP (Optional[np.ndarray], optional): Spatial predictor matrix, n*sp matrix. Defaults to None.
            train (bool, optional): train or evaluation mode. Defaults to False.
            batch_size (int, optional): Batch size. Defaults to 25.
            parallel (int, optional): Parallelization of DataLoader. Defaults to 0.
            sampling (int, optional): Number of MC-samples for each species. Defaults to 100.
            link (bool, optional): Linear or response scale. Defaults to True.

        Returns:
            np.ndarray: Predictions
        """                

        dataLoader = self._get_DataLoader(X = newdata, Y = None, SP=SP, batch_size = batch_size, shuffle = False, parallel = parallel, drop_last = False)
        loss_function = self._build_loss_function(train = False, raw=not link)

        pred = []
        if self.device.type == 'cuda':
            device = self.device.type+ ":" + str(self.device.index)
        else:
            device = 'cpu'
        
        if type(SP) is np.ndarray:
            for step, (x, sp) in enumerate(dataLoader):
                x = x.to(self.device, non_blocking=True)
                sp = sp.to(self.device, non_blocking=True)
                mu = self.env(x) + self.spatial(sp)
                # loss = self._loss_function(mu, y, self.sigma, batch_size, sampling, df, alpha, device)
                loss = loss_function(mu, self.sigma, x.shape[0], sampling, self.df, self.alpha, device)
                pred.append(loss)
        else:
            for step, (x) in enumerate(dataLoader):
                x = x[0].to(self.device, non_blocking=True)
                mu = self.env(x)
                # loss = self._loss_function(mu, y, self.sigma, batch_size, sampling, df, alpha, device)
                loss = loss_function(mu, self.sigma, x.shape[0], sampling, self.df, self.alpha, device)
                pred.append(loss)
        return torch.cat(pred, dim = 0).data.cpu().numpy()

    def se(self, 
           X: np.ndarray, 
           Y: np.ndarray, 
           SP: Optional[np.ndarray] = None, 
           batch_size: int = 25, 
           parallel: int = 0, 
           sampling: int = 100) -> List[np.ndarray]:
        """Calculate standard errors for environmental coefficients

        Args:
            X (np.ndarray): Environment, n*p matrix
            Y (np.ndarray): Species occurences, n*s matrix
            SP (Optional[np.ndarray], optional): Spatial predictor matrix, n*sp matrix. Defaults to None.
            batch_size (int, optional): Batch size. Defaults to 25.
            parallel (int, optional): Parallelization of DataLoader. Defaults to 0.
            sampling (int, optional): Number of MC samples for each species. Defaults to 100.

        Returns:
            List[np.ndarray]: Standard errors for environmental coefficients
        """           
        dataLoader = self._get_DataLoader(X, Y, SP, batch_size=batch_size, shuffle=False)
        loss_func = self._build_loss_function(train=True)
        se = []
        weights_base = np.transpose(self.env_weights[0])
        y_dim = Y.shape[1]

        if self.device.type == 'cuda':
            device = self.device.type+ ":" + str(self.device.index)
        else:
            device = 'cpu'
        
        _ = sys.stdout.write("\nCalculating standard errors...\n")

        desc='loss: Inf'
        sp_bar = tqdm(range(Y.shape[1]),bar_format= "Species: {n_fmt}/{total_fmt} {l_bar}{bar}| [{elapsed}, {rate_fmt}]", file=sys.stdout)

        if type(SP) is np.ndarray:
            for i in sp_bar:
                weights = torch.tensor(weights_base[:,i].reshape([-1,1]), device=self.device, dtype=self.dtype, requires_grad=True).to(self.device)
                if i == 0:
                    constants = torch.tensor(weights_base[:,(i+1):], device=self.device, dtype=self.dtype).to(self.device)
                    w = torch.cat([weights, constants], dim=1)
                elif i < y_dim:
                    w = torch.cat([torch.tensor(weights_base[:,0:i], device=self.device, dtype=self.dtype).to(self.device), 
                                   weights, 
                                   torch.tensor(weights_base[:,(i+1):], device=self.device, dtype=self.dtype).to(self.device)],dim=1)
                else:
                    constants = torch.tensor(weights_base[:,0:i], device=self.device, dtype=self.dtype).to(self.device)
                    w = torch.cat([constants, weights], dim=1)
                for step, (x, y, sp) in enumerate(dataLoader):
                    x = x.to(self.device, non_blocking=True)
                    y = y.to(self.device, non_blocking=True)
                    sp = sp.to(self.device, non_blocking=True)
                    mu = torch.nn.functional.linear(x, w.t()) + self.spatial(sp)
                    loss = loss_func(mu, y,self.sigma, x.shape[0], sampling, self.df, self.alpha, device, self.dtype).sum()
                    first_gradients = torch.autograd.grad(loss, weights, retain_graph=True, create_graph=True, allow_unused=True)
                    second = []
                    for j in range(self.input_shape):
                        second.append(torch.autograd.grad(first_gradients[0][j,0], inputs = weights, retain_graph = True, create_graph = False, allow_unused = False)[0])
                        hessian = torch.cat(second,dim=1)
                    if step < 1:
                        hessian_out = hessian
                    else:
                        hessian_out += hessian
                se.append(torch.sqrt(torch.diag(torch.inverse(hessian_out))).data.cpu().numpy())
            return se                
        else:
            for i in sp_bar:
                weights = torch.tensor(weights_base[:,i].reshape([-1,1]), device=self.device, dtype=self.dtype, requires_grad=True).to(self.device)
                if i == 0:
                    constants = torch.tensor(weights_base[:,(i+1):], device=self.device, dtype=self.dtype).to(self.device)
                    w = torch.cat([weights, constants], dim=1)
                elif i < y_dim:
                    w = torch.cat([torch.tensor(weights_base[:,0:i], device=self.device, dtype=self.dtype).to(self.device), 
                                   weights, 
                                   torch.tensor(weights_base[:,(i+1):], device=self.device, dtype=self.dtype).to(self.device)],dim=1)
                else:
                    constants = torch.tensor(weights_base[:,0:i], device=self.device, dtype=self.dtype).to(self.device)
                    w = torch.cat([constants, weights], dim=1)
                for step, (x, y) in enumerate(dataLoader):
                    x = x.to(self.device, non_blocking=True)
                    y = y.to(self.device, non_blocking=True)
                    mu = torch.nn.functional.linear(x, w.t())
                    loss = loss_func(mu, y, self.sigma, x.shape[0], sampling, self.df, self.alpha, device, self.dtype).sum()
                    first_gradients = torch.autograd.grad(loss, weights, retain_graph=True, create_graph=True, allow_unused=True)
                    second = []
                    for j in range(self.input_shape):
                        second.append(torch.autograd.grad(first_gradients[0][j,0], inputs = weights, retain_graph = True, create_graph = False, allow_unused = False)[0])
                        hessian = torch.cat(second,dim=1)
                    if step < 1:
                        hessian_out = hessian
                    else:
                        hessian_out += hessian
                se.append(torch.sqrt(torch.diag(torch.inverse(hessian_out))).data.cpu().numpy())
            return se

    def _build_cov_constrain_function(self, 
                                      l1: float = None, 
                                      l2: float = None, 
                                      reg_on_Cov: Optional[bool] = None, 
                                      reg_on_Diag: Optional[bool] = None, 
                                      inverse: Optional[bool] = None) -> None:
        """Build regularization losses for species associations

        Args:
            l1 (float, optional): LASSO regularization. Defaults to None.
            l2 (float, optional): Ridge regularization. Defaults to None.
            reg_on_Cov (Optional[bool], optional): Regularization on covariance matrix. Defaults to None.
            reg_on_Diag (Optional[bool], optional): Regularization on diagonals. Defaults to None.
            inverse (Optional[bool], optional): Regularization on precision matrix. Defaults to None.

        """                                      
        if reg_on_Cov:
            if reg_on_Diag:
                diag = int(0)
            else:
                diag = int(1)
            
            if l1 > 0.0 and l2 > 0.0:
                if inverse:
                    identity = torch.eye(self.sigma.shape[0], dtype=self.sigma.dtype, device=self.sigma.device).to(self.sigma.device)
                    @torch.jit.script
                    def l1_l2_ll(sigma: torch.Tensor, l1: float, l2: float, diag: int, identity: torch.Tensor):
                        sigma1 = sigma.matmul(sigma.t()).add(identity)
                        #ss = sigma1.add(identity).inverse()
                        ss = identity.cholesky_solve(sigma1.cholesky())
                        return ss.triu(diag).abs().sum().mul(l1) + ss.tril(-1).abs().sum().mul(l1) + ss.triu(diag).pow(2.0).sum().mul(l2) + ss.tril(-1).pow(2.0).sum().mul(l2) #+ sigma1.pow(2.0).sum().mul(0.0001)
                    self.losses.append(lambda: l1_l2_ll(self.sigma, l1,l2, diag, identity))
                else:
                    @torch.jit.script
                    def l1_l2_ll(sigma: torch.Tensor, l1: float, l2: float, diag: int):
                        ss = sigma.matmul(sigma.t())
                        #ss = ss.add(identity).inverse()
                        return ss.triu(diag).abs().sum().mul(l1) + ss.tril(-1).abs().sum().mul(l1) + ss.triu(diag).pow(2.0).sum().mul(l2) + ss.tril(-1).pow(2.0).sum().mul(l2)
                    self.losses.append(lambda: l1_l2_ll(self.sigma, l1,l2, diag))
            else:    
                if l1 > 0.0:
                    if inverse:
                        identity = torch.eye(self.sigma.shape[0], dtype=self.sigma.dtype, device=self.sigma.device).to(self.sigma.device)
                        @torch.jit.script
                        def l1_ll(sigma: torch.Tensor, l1: float, diag: int, identity: torch.Tensor):
                            sigma2= sigma.matmul(sigma.t()).add(identity)
                            #ss = sigma2.add(identity).inverse()
                            ss = identity.cholesky_solve(sigma2.cholesky())
                            return ss.triu(diag).abs().sum().mul(l1) + ss.tril(-1).abs().sum().mul(l1) #  + sigma2.pow(2.0).sum().mul(0.0001)
                        self.losses.append(lambda: l1_ll(self.sigma, l1, diag, identity))
                    else:
                        @torch.jit.script
                        def l1_ll(sigma: torch.Tensor, l1: float, diag: int):
                            ss = sigma.matmul(sigma.t())
                            return ss.triu(diag).abs().sum().mul(l1) + ss.tril(-1).abs().sum().mul(l1)
                        self.losses.append(lambda: l1_ll(self.sigma, l1, diag))                    

                if l2 > 0.0 :
                    if inverse:
                        identity = torch.eye(self.sigma.shape[0], dtype=self.sigma.dtype, device=self.sigma.device).to(self.sigma.device)
                        @torch.jit.script
                        def l2_ll(sigma: torch.Tensor, l2: float, diag: int, identity: torch.Tensor):
                            sigma2 = sigma.matmul(sigma.t()).add(identity)
                            #ss = sigma2.inverse()
                            ss = identity.cholesky_solve(sigma2.cholesky())
                            return ss.triu(diag).pow(2.0).sum().mul(l2) + ss.tril(-1).pow(2.0).sum().mul(l2)
                        self.losses.append(lambda: l2_ll(self.sigma, l2, diag, identity))
                    else:
                        identity = torch.eye(self.sigma.shape[0], dtype=self.sigma.dtype, device=self.sigma.device).to(self.sigma.device)
                        @torch.jit.script
                        def l2_ll(sigma: torch.Tensor, l2: float, diag: int):
                            ss = sigma.matmul(sigma.t())
                            return ss.triu(diag).pow(2.0).sum().mul(l2) + ss.tril(-1).pow(2.0).sum().mul(l2)
                        self.losses.append(lambda: l2_ll(self.sigma, l2, diag))                    
        else:
            if l1 > 0.0:
                self.losses.append( lambda: self.l1_l2[0](self.sigma, l1) )
            if l2 > 0.0:
                self.losses.append( lambda: self.l1_l2[1](self.sigma, l2) )

    def _build_loss_function(self, train: bool = True, raw: bool = False) -> Callable:
        """Build loss (likelihood) function

        Args:
            train (bool, optional): Train or evaulation mode Defaults to True.
            raw (bool, optional): Linear or response scale. Defaults to False.

        Returns:
            Callable: loss function
        """        

        if train:
            if self.link == "logit" or self.link == "probit":
                @torch.jit.script
                def tmp(mu: torch.Tensor, Ys: torch.Tensor, sigma: torch.Tensor, batch_size: int, sampling: int, df: int, alpha: float, device: str, dtype: torch.dtype):
                    noise = torch.randn(size = [sampling, batch_size, df], device=torch.device(device), dtype=dtype)
                    E = torch.sigmoid(   torch.tensordot(noise, sigma.t(), [2], [0]).add(mu).mul(alpha)   ).mul(0.999999).add(0.0000005)
                    logprob = E.log().mul(Ys).add((1.0 - E).log().mul(1.0 - Ys)).neg().sum(dim = 2).neg()
                    maxlogprob = logprob.max(dim = 0).values
                    Eprob = logprob.sub(maxlogprob).exp().mean(dim = 0)
                    loss = Eprob.log().neg().sub(maxlogprob)
                    return loss
            elif self.link == "linear":
                @torch.jit.script
                def tmp(mu: torch.Tensor, Ys: torch.Tensor, sigma: torch.Tensor, batch_size: int, sampling: int, df: int, alpha: float, device: str, dtype: torch.dtype):
                    noise = torch.randn(size = [sampling, batch_size, df], device=torch.device(device), dtype=dtype)
                    E = torch.clamp(torch.tensordot(noise, sigma.t(), [2], [0]).add(mu).mul(alpha), 0.0, 1.0).mul(0.999999).add(0.0000005)
                    logprob = E.log().mul(Ys).add((1.0 - E).log().mul(1.0 - Ys)).neg().sum(dim = 2).neg()
                    maxlogprob = logprob.max(dim = 0).values
                    Eprob = logprob.sub(maxlogprob).exp().mean(dim = 0)
                    loss = Eprob.log().neg().sub(maxlogprob)
                    return loss
            elif self.link == "count":
                def tmp(mu: torch.Tensor, Ys: torch.Tensor, sigma: torch.Tensor, batch_size: int, sampling: int, df: int, alpha: float, device: str, dtype: torch.dtype):
                    noise = torch.randn(size = [sampling, batch_size, df], device=torch.device(device),dtype=dtype)
                    E = torch.tensordot(noise, sigma.t(), dims = 1).add(mu).exp()
                    logprob = torch.distributions.Poisson(rate=E).log_prob(Ys).sum(2)
                    maxlogprob = logprob.max(dim = 0).values
                    Eprob = logprob.sub(maxlogprob).exp().mean(dim = 0)
                    return Eprob.log().neg().sub(maxlogprob)

            elif self.link == "normal":
                def tmp(mu: torch.Tensor, Ys: torch.Tensor, sigma: torch.Tensor, batch_size: int, sampling: int, df: int, alpha: float, device: str, dtype: torch.dtype):
                    return torch.distributions.MultivariateNormal(loc=mu, covariance_matrix=sigma.matmul(sigma.t()).add(torch.eye(sigma.shape[0]))).log_prob(Ys).neg()

        else:
            if self.link == "probit": 
                link_func = lambda value: torch.distributions.Normal(0.0, 1.0).cdf(value)
            elif self.link == "logit":
                link_func = lambda value: torch.sigmoid(value)
            elif self.link == "linear":
                link_func = lambda value: torch.clamp(value, 0.0, 1.0)
            elif self.link == "count":
                link_func = lambda value: value.exp()
            elif self.link == "normal":
                link_func = lambda value: value

            if raw:
                link_func = lambda value: value

            def tmp(mu: torch.Tensor, sigma: torch.Tensor, batch_size: int, sampling: int, df: int, alpha: float, device: str):
                noise = torch.randn(size = [sampling, batch_size, df], device=torch.device(device))
                if self.link == "logit": 
                    E = link_func(torch.tensordot(noise, sigma.t(), 1).add(mu).mul(alpha)).mul(0.999999).add(0.0000005)
                else:
                    E = link_func(torch.tensordot(noise, sigma.t(), 1).add(mu)).mul(0.999999).add(0.0000005)
                return E.mean(dim = 0)

        return tmp

    @property
    def weights(self):
        return [(lambda p: p.data.cpu().numpy())(p) for p in self.params()] 

    @property
    def get_sigma(self):
        return self.sigma.data.cpu().numpy()

    def set_sigma(self, w: np.ndarray):
        """Set sigma

        Args:
            w (np.ndarray): Square-root sigma, s*df matrix 
        """        
        with torch.no_grad():
            self.sigma.data = torch.tensor(w, device=self.device, dtype=self.dtype).data
            
    @property
    def covariance(self):
        return (self.sigma.matmul(self.sigma.t()) + torch.eye(self.sigma.shape[0], dtype=self.sigma.dtype, device=self.device)).data.cpu().numpy()

    @property
    def env_weights(self):
        return [(lambda p: p.data.cpu().numpy())(p) for p in self.env.parameters()]

    def set_env_weights(self, w: np.ndarray):
        """Set environmental coefficients

        Args:
            w (np.ndarray): environmental coefficients, p*s matrix
        """        
        with torch.no_grad():
            counter = 0
            for i in range(len(self.env)):
                if type(self.env[i]) is torch.nn.modules.linear.Linear:
                    self.env[i].weight = torch.nn.Parameter(torch.tensor(w[counter], dtype=self.env[i].weight.dtype, device=self.env[i].weight.device))
                    counter+=1
                    if self.env[i].bias is not None:
                        self.env[i].bias = torch.nn.Parameter(torch.tensor(w[counter], dtype=self.env[i].bias.dtype, device=self.env[i].bias.device))
                        counter+=1

    @property
    def spatial_weights(self):
        if self.spatial is not None:
            return [(lambda p: p.data.cpu().numpy())(p) for p in self.spatial.parameters()]
        else:
            return None

    def set_spatial_weights(self, w: np.ndarray):
        """Set spatial coefficients

        Args:
            w (np.ndarray): spatial coefficients, sp*ps matrix

        """        
        if self.spatial is None:
            return None
        with torch.no_grad():
            counter = 0
            for i in range(len(self.spatial)):
                if type(self.spatial[i]) is torch.nn.modules.linear.Linear:
                    self.spatial[i].weight = torch.nn.Parameter(torch.tensor(w[counter], dtype=self.spatial[i].weight.dtype, device=self.spatial[i].weight.device))
                    counter+=1
                    if self.spatial[i].bias is not None:
                        self.spatial[i].bias = torch.nn.Parameter(torch.tensor(w[counter], dtype=self.spatial[i].bias.dtype, device=self.spatial[i].bias.device))
                        counter+=1



import pytest
from .. import sjSDM_py as fa
import numpy as np

@pytest.fixture
def data():
    def _get(a=5, b=5, c=100):
        return (np.random.randn(c,a), np.random.binomial(1,0.5,[c, b]))
    return _get

@pytest.fixture
def data_sp():
    def _get(a=5, b=5, c=100):
        return (np.random.randn(c,a), np.random.binomial(1,0.5,[c, b]), np.random.uniform(-1,1,[c,2]))
    return _get


@pytest.fixture
def model_base():
    def _get(inp=5,out=5,hidden=[], activation=['linear'],bias=[False], df=5, 
             optimizer=fa.optimizer_adamax(1), l1_d = 0.0, l2_d = 0.0, 
             l1_cov=0.0, l2_cov=0.0, link="logit",reg_on_Cov=True, reg_on_Diag=True,inverse=False, dropout=-99):
        model = fa.Model_sjSDM()
        model.add_env(input_shape=inp, output_shape=out, bias=bias,
                      hidden=hidden, activation=activation, l1=l1_d, l2=l2_d, dropout=dropout)
        model.build(df=df, l1=l1_cov,l2=l2_cov,optimizer=optimizer,link=link,reg_on_Cov=reg_on_Cov, reg_on_Diag=reg_on_Diag, inverse=inverse)
        return model
    return _get

@pytest.fixture
def model_base_sp():
    def _get(inp=5,out=5,hidden=[], activation=['linear'],bias=[False], df=5, 
            hidden_sp=[],activation_sp=['linear'],bias_sp=[False], sp_l1=0.0, sp_l2=0.0, 
             optimizer=fa.optimizer_adamax(1), l1_d = 0.0, l2_d = 0.0, 
             l1_cov=0.0, l2_cov=0.0, link="logit",reg_on_Cov=True, reg_on_Diag=True,inverse=False):
        model = fa.Model_sjSDM()
        model.add_env(input_shape=inp, output_shape=out, bias=bias,
                      hidden=hidden, activation=activation, l1=l1_d, l2=l2_d)
        model.add_spatial(input_shape=2, output_shape=out, hidden=hidden_sp,bias=bias_sp, activation=activation_sp,l1=sp_l1,l2=sp_l2)
        model.build(df=df, l1=l1_cov,l2=l2_cov,optimizer=optimizer,link=link,reg_on_Cov=reg_on_Cov, reg_on_Diag=reg_on_Diag, inverse=inverse)
        return model
    return _get


@pytest.mark.parametrize("inp,out,epochs,batch,early_stopping_training", 
                        [(5,5,1,10,-1), 
                         (3,10,2,2,-1),
                         (1,15,5,5,2),
                         (20,11,1, 100, 2),
                         pytest.param(5,5,-1,1,-1, marks=pytest.mark.xfail),
                         pytest.param(5,5,2,-1,-1, marks=pytest.mark.xfail),
                         pytest.param(0,5,2,1,-1, marks=pytest.mark.xfail),
                         pytest.param(5,0,2,1,-1, marks=pytest.mark.xfail)])
def test_base(data, model_base, inp, out, epochs, batch, early_stopping_training):
    X, Y = data(inp, out)
    model = model_base(inp, out)
    model.fit(X, Y, epochs=epochs, batch_size=batch, early_stopping_training=early_stopping_training)
    model.logLik(X, Y, batch_size=batch)
    pred = model.predict(X, batch_size=batch)
    model.se(X, Y, batch_size=batch)
    assert pred.shape[0] == X.shape[0]


@pytest.mark.parametrize("links", [("probit"), ("linear"), ("logit"), pytest.param("failed", marks=pytest.mark.xfail)])
def test_link(data, model_base, links):
    X, Y = data()
    model = model_base(link=links)
    model.fit(X, Y, epochs=1, batch_size=50)
    assert model.link == links
    model.logLik(X, Y)
    pred = model.predict(X)
    model.se(X, Y)
    assert pred.shape[0] == X.shape[0]


@pytest.mark.parametrize("l1_d,l2_d,l2_cov,l1_cov,reg_on_Cov,reg_on_Diag,inverse", 
                        [
                            (-1,0.1, -1, -1, True, True, False),
                            (0.1,-1, -1, -1, True, True, False),
                            (0.1,0.1, -1, -1, True, True, False),
                            (0.1,0.1, -1, 0.1, True, True, False),
                            (0.1,0.1, 0.1, -1, True, True, False),
                            (0.1,0.1, 0.1, 0.1, True, True, False),
                            (0.1,0.1, 0.1, 0.1, False, True, False),
                            (0.1,0.1, 0.1, 0.1, True, False, False),
                            (0.1,0.1, 0.1, 0.1, True, False, True)
                        ])
def test_regularization(data, model_base, l1_d, l2_d, l2_cov, l1_cov, reg_on_Cov, reg_on_Diag,inverse):
    X, Y = data()
    model = model_base(5, 5, l1_d=l1_d, l2_d=l2_d, l2_cov=l2_cov, l1_cov=l1_cov, reg_on_Cov=reg_on_Cov, reg_on_Diag=reg_on_Diag, inverse=inverse)
    model.fit(X, Y, epochs=1, batch_size=10)
    model.logLik(X, Y, batch_size=10)
    pred = model.predict(X, batch_size=10)
    model.se(X, Y)
    assert pred.shape[0] == X.shape[0]

@pytest.mark.parametrize("inp,out,hidden,activation,bias,dropout", 
                        [
                            (5,5,[], ['linear'],[False], 0.1),
                            (3,3,[5,5,5], ['linear'],[True], 0.1),
                            (2,10,[3,3,3], ['relu'], [True,False,True,True], -1),
                            (2,10,[3,3,3], ['sigmoid'],[True], -1),
                            (2,10,[3,3,3], ['tanh'], [True], -1),
                            (2,10,[3,3,3], ['selu'], [True], -1),
                            (2,10,[3,3,3], ['leakyrelu'], [True], -1),
                            (2,10,[3,3,3], ['relu', 'tanh', 'sigmoid'],[True, False, True, True], 0.3)
                        ])
def test_dnn(data, model_base, inp, out, hidden, activation,bias, dropout):
    X, Y = data(inp, out)
    model = model_base(inp, out, hidden=hidden, activation=activation, dropout=dropout)
    model.fit(X, Y, epochs=1, batch_size=50)
    model.logLik(X, Y)
    pred = model.predict(X)
    assert pred.shape[0] == X.shape[0]



@pytest.mark.parametrize("inp,out,epochs,batch", 
                        [(5,5,1,10), 
                         (3,10,2,2),
                         (1,15,5,5),
                         (20,11,1, 100),
                         pytest.param(5,5,-1,1, marks=pytest.mark.xfail),
                         pytest.param(5,5,2,-1, marks=pytest.mark.xfail),
                         pytest.param(0,5,2,1, marks=pytest.mark.xfail),
                         pytest.param(5,0,2,1, marks=pytest.mark.xfail)])
def test_base_sp(data_sp, model_base_sp, inp, out, epochs, batch):
    X, Y, SP = data_sp(inp, out)
    model = model_base_sp(inp, out)
    model.fit(X, Y, epochs=epochs, batch_size=batch)
    model.logLik(X, Y,SP=SP, batch_size=batch)
    pred = model.predict(X, SP=SP, batch_size=batch)
    model.se(X, Y, SP=SP, batch_size=batch)
    assert pred.shape[0] == X.shape[0]


@pytest.mark.parametrize("inp,out,hidden,activation,l1,l2", 
                        [
                            (5,5,[], ['linear'], -1,-1),
                            (3,3,[5,5,5], ['linear'],-1,-1),
                            (2,10,[3,3,3], ['relu'],0.1,-1),
                            (2,10,[3,3,3], ['sigmoid'], -1, 0.1),
                            (2,10,[3,3,3], ['tanh'], 0.1, 0.1),
                            (2,10,[3,3,3], ['relu', 'tanh', 'sigmoid'], 0.5, 0.5)
                        ])
def test_dnn_sp(data_sp, model_base_sp, inp, out, hidden, activation,l1,l2):
    X, Y, SP = data_sp(inp, out)
    model = model_base_sp(inp, out, hidden_sp=hidden, activation_sp=activation,sp_l1=l1,sp_l2=l2)
    model.fit(X, Y, SP=SP, epochs=1, batch_size=50)
    model.logLik(X, Y, SP=SP)
    pred = model.predict(X, SP=SP)
    assert pred.shape[0] == X.shape[0]


def test_base_weights(data_sp, model_base_sp):
    epochs = 1
    batch = 5
    X, Y, SP = data_sp()
    model = model_base_sp(5, 5, hidden = [5,5,5], hidden_sp=[3,3,3])
    model.fit(X, Y, epochs=epochs, batch_size=batch)

    model.set_env_weights(model.env_weights)
    model.set_spatial_weights(model.spatial_weights)

    ## still buggy
    #model.logLik(X, Y,SP=SP, batch_size=batch)
    #pred = model.predict(X, SP=SP, batch_size=batch)
    #model.se(X, Y, SP=SP, batch_size=batch)
    #assert pred.shape[0] == X.shape[0]


from .dist_mvp import MultivariateProbit
import torch
import numpy as np
import pyro
import logging
import sys
logging.basicConfig(format='%(message)s', level=logging.INFO)

class Model_SVI():
    def __init__(self, device="cpu", dtype="float32"):
        device, dtype = self._device_and_dtype(device, dtype)
        self.device = device
        self.dtype = dtype
    
    @staticmethod
    def _device_and_dtype(device, dtype):
        """define device and dtype

        parse device and dtype into torch.dtype and torch.device

        # Arguments
            device: str or torch.device, "cpu" or "gpu"
            dtype: str or torch.dtype, "float32" and "float64" are supported

        # Example

        ```python
            device, dtype = _device_and_dtype(torch.device("cpu"), "float32")
            device, dtype = _device_and_dtype("gpu", "float32")
        ````

        """
        if type(device) is int:
            device = torch.device('cuda:'+str(device))

        if type(device) is str:
            device = torch.device(device)

        if type(dtype) is not str:
            return device, dtype
        else:
            if dtype == "float32":
                return device, torch.float32
            if dtype == "float64":
                return device, torch.float64
        return device, torch.float32
    
    def _get_DataLoader(self, X, Y=None,SP=None,RE=None, batch_size=25, shuffle=True, parallel=0, drop_last=True):
        if self.device.type == 'cuda':
            torch.cuda.set_device(self.device)
            pin_memory = False
        else:
            pin_memory = True
        #init_func = lambda: torch.multiprocessing.set_start_method('spawn', True)
        if type(RE) is np.ndarray:
            if type(Y) is np.ndarray:
                if type(SP) is np.ndarray:
                    data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=torch.float32, device=torch.device('cpu')), 
                                                          torch.tensor(Y, dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(SP, dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.asarray(RE).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))
                else:
                    data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=torch.float32, device=torch.device('cpu')), 
                                                          torch.tensor(Y, dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.asarray(RE).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))

            else:
                if type(SP) is np.ndarray:
                    data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(SP, dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.asarray(RE).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))
                else: 
                    data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=torch.float32, device=torch.device('cpu')),
                                                          torch.tensor(np.asarray(RE).reshape([-1,1]), dtype=torch.long, device=torch.device('cpu')))
        
        else: 
            if type(Y) is np.ndarray:
                if type(SP) is np.ndarray:
                    data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=torch.float32, device=torch.device('cpu')), 
                                                        torch.tensor(Y, dtype=torch.float32, device=torch.device('cpu')),
                                                        torch.tensor(SP, dtype=torch.float32, device=torch.device('cpu')))
                else:
                    data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=torch.float32, device=torch.device('cpu')), 
                                                        torch.tensor(Y, dtype=torch.float32, device=torch.device('cpu')))

            else:
                if type(SP) is np.ndarray:
                    data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=torch.float32, device=torch.device('cpu')),
                                                        torch.tensor(SP, dtype=torch.float32, device=torch.device('cpu')))
                else: 
                    data = torch.utils.data.TensorDataset(torch.tensor(X, dtype=torch.float32, device=torch.device('cpu')))            

        DataLoader = torch.utils.data.DataLoader(data, batch_size=batch_size, shuffle=shuffle, num_workers=int(parallel), pin_memory=pin_memory, drop_last=drop_last)
        torch.cuda.empty_cache()
        return DataLoader
    
    def build_model(self, X, Y,Re,SP, df, scale_mu = 5.0, scale_scale = 1.0, batch_size = 20):
        pyro.enable_validation(True)
        pyro.clear_param_store()
        #XX = torch.tensor(X, dtype = torch.float32)
        #YY = torch.tensor(Y, dtype = torch.float32)
        self.e = X.shape[1]
        self.sp = Y.shape[1]
        self.df = df
        e = self.e
        sp = self.sp

        if Re is None:
            def model(XX, YY=None, link = self.link):
                mu_scale = torch.ones([e, sp])
                mu_loc = torch.zeros([e, sp])
                scale_scale2 = torch.ones([sp, df])
                scale_loc = torch.zeros([sp, df])
                mu = pyro.sample("mu", pyro.distributions.Normal(mu_loc, mu_scale).to_event())
                scale = pyro.sample("scale", pyro.distributions.Normal(scale_loc, scale_scale2).to_event())
                with pyro.plate('data', size = XX.shape[0]):
                    loc_tmp = XX.matmul(mu)
                    #loc_tmp = mu(XX[ind,:])
                    pyro.sample("obs", MultivariateProbit(loc_tmp, scale), obs = YY)
        else:
            len_unique = np.unique(Re[:,0]).shape[0]
            #indices = torch.tensor(Re, dtype=torch.long)
            def model(XX, YY=None, Re=None, link = self.link):
                mu_scale = torch.ones([e, sp])*scale_mu
                mu_loc = torch.zeros([e, sp])
                scale_scale2 = torch.ones([sp, df])*scale_scale
                scale_loc = torch.zeros([sp, df])
                mu = pyro.sample("mu", pyro.distributions.Normal(mu_loc, mu_scale).to_event())
                scale = pyro.sample("scale", pyro.distributions.Normal(scale_loc, scale_scale2).to_event())
                re = pyro.sample("re", pyro.distributions.Normal(0.0, torch.ones([len_unique,1])).to_event())
                with pyro.plate('data', size = XX.shape[0]):
                    if Re is None:
                        loc_tmp = XX.matmul(mu)
                    else:
                        loc_tmp = XX.matmul(mu) + re.gather(0, Re)
                    pyro.sample("obs", MultivariateProbit(loc_tmp, scale), obs = YY)
        
        return model

    def get_guide(self):
        df = self.df
        e = self.e
        sp = self.sp
        if Re is None:
            def guide(XX, YY):
                mu_loc_0 = pyro.param("mu_loc_0", torch.zeros([e, sp]))
                mu_scale_0 = pyro.param("mu_scale_0", torch.ones([e, sp])*7.0, constraint=pyro.distributions.constraints.positive)
                scale_loc_0 = pyro.param("scale_loc_0", torch.zeros([sp, df]))
                scale_scale_0 = pyro.param("scale_scale_0", torch.ones([sp, df])*1.0, constraint=pyro.distributions.constraints.positive)
                mu = pyro.sample("mu", pyro.distributions.Normal(mu_loc_0, mu_scale_0).to_event()) #, infer = baseline_dict)
                scale = pyro.sample("scale", pyro.distributions.Normal(scale_loc_0, scale_scale_0).to_event()) #, infer = baseline_dict)

            return guide
        else:
            def guide(XX, YY, Re):
                mu_loc_0 = pyro.param("mu_loc_0", torch.zeros([e, sp]))
                mu_scale_0 = pyro.param("mu_scale_0", torch.ones([e, sp])*7.0, constraint=pyro.distributions.constraints.positive)
                scale_loc_0 = pyro.param("scale_loc_0", torch.zeros([sp, df]))
                scale_scale_0 = pyro.param("scale_scale_0", torch.ones([sp, df])*1.0, constraint=pyro.distributions.constraints.positive)
                mu = pyro.sample("mu", pyro.distributions.Normal(mu_loc_0, mu_scale_0).to_event()) #, infer = baseline_dict)
                scale = pyro.sample("scale", pyro.distributions.Normal(scale_loc_0, scale_scale_0).to_event()) #, infer = baseline_dict)

            return guide

    def fit(self, X, Y, df,RE=None, SP=None,
            guide='LaplaceApproximation', 
            scale_mu = 5.0, scale_scale = 0.01, 
            lr = 0.001, epochs = 200,
            batch_size = 25,
            num_samples=100,
            parallel = 0,
            link = "logit"):
        
        if self.device.type == 'cuda':
            torch.set_default_tensor_type('torch.cuda.FloatTensor')
        else:
            torch.set_default_tensor_type('torch.FloatTensor')
            

        self.link = link
        torch.cuda.empty_cache()
        guide2 = guide
        self.model = self.build_model(X, Y,RE,SP, df, scale_mu, scale_scale, batch_size)
        elbo = pyro.infer.JitTrace_ELBO(ignore_jit_warnings=True)
        if guide2 == 'LaplaceApproximation':
            self.guide = pyro.infer.autoguide.AutoLaplaceApproximation(self.model)

        if guide2 == 'LowRankMultivariateNormal':
            #self.guide = pyro.infer.autoguide.AutoLowRankMultivariateNormal(self.model, init_loc_fn=pyro.infer.autoguide.init_to_mean)
            self.guide = pyro.infer.autoguide.AutoGuideList(self.model)
            self.guide.add(pyro.infer.autoguide.AutoLowRankMultivariateNormal(pyro.poutine.block(self.model, expose=["mu"]), init_loc_fn=pyro.infer.autoguide.init_to_mean, init_scale=scale_mu))
            self.guide.add(pyro.infer.autoguide.AutoLowRankMultivariateNormal(pyro.poutine.block(self.model, expose=["scale"]),  init_loc_fn=pyro.infer.autoguide.init_to_mean, init_scale=scale_scale))
            elbo = pyro.infer.Trace_ELBO(ignore_jit_warnings=True)
            if type(RE) is np.ndarray:
                self.guide.add(pyro.infer.autoguide.AutoLowRankMultivariateNormal(pyro.poutine.block(self.model, expose=["re"]), init_loc_fn=pyro.infer.autoguide.init_to_mean, init_scale=1.0))

        if guide2 == "Delta":
            self.guide = pyro.infer.autoguide.AutoDelta(self.model, init_loc_fn=pyro.infer.autoguide.init_to_mean)

        if guide2 == 'DiagonalNormal':
            self.guide = pyro.infer.autoguide.AutoGuideList(self.model)
            self.guide.add(pyro.infer.autoguide.AutoDiagonalNormal(pyro.poutine.block(self.model, expose=["mu"]), init_loc_fn=pyro.infer.autoguide.init_to_mean, init_scale=scale_mu ))
            self.guide.add(pyro.infer.autoguide.AutoDiagonalNormal(pyro.poutine.block(self.model, expose=["scale"]),  init_loc_fn=pyro.infer.autoguide.init_to_mean, init_scale=scale_scale))
            if type(RE) is np.ndarray:
                self.guide.add(pyro.infer.autoguide.AutoDiagonalNormal(pyro.poutine.block(self.model, expose=["re"]), init_loc_fn=pyro.infer.autoguide.init_to_mean, init_scale=1.0))


        if guide2 == 'manually':
            self.guide = self.get_guide()
        
        if len(lr) == 1:
             adam = pyro.optim.Adam({'lr' : lr[0]})
        else:
            def per_param_callable(module_name, param_name):
                if param_name == 'mu':
                    return {"lr": lr[0]}
                else:
                    return {"lr": lr[1]}
            adam = pyro.optim.Adam(per_param_callable)
        #elbo = pyro.infer.Trace_ELBO(ignore_jit_warnings=True)
        #guide()
        #XX = torch.tensor(X, dtype=torch.float32)
        #YY = torch.tensor(Y, dtype=torch.float32)
        
        self.svi = pyro.infer.SVI(self.model, self.guide, adam, loss = elbo)
        stepSize = np.floor(X.shape[0] / batch_size).astype(int)
        dataLoader = self._get_DataLoader(X, Y, SP, RE, batch_size, True, parallel)
        batch_loss = np.zeros(stepSize)
        self.history = np.zeros(epochs)
        if RE is None:          
            for epoch in range(epochs):
                for step, (x, y) in enumerate(dataLoader):
                    x = x.to(self.device, non_blocking=True)
                    y = y.to(self.device, non_blocking=True)
                    loss = self.svi.step(x, y)
                    batch_loss[step] = loss
                bl = np.mean(batch_loss)
                _ = sys.stdout.write("\rEpoch: {}/{} loss: {} ".format(epoch+1,epochs, np.round(bl, 3).astype(str)))
                sys.stdout.flush()
            self.posterior_samples = pyro.infer.Predictive(self.model, guide=self.guide, num_samples=num_samples)(torch.tensor(X, dtype=torch.float32), 
                                                                                                                  torch.tensor(Y, dtype=torch.float32))

        else:
            for epoch in range(epochs):
                for step, (x, y, re) in enumerate(dataLoader):
                    x = x.to(self.device, non_blocking=True)
                    y = y.to(self.device, non_blocking=True)
                    re=re.to(self.device, non_blocking=True)
                    #print(re.dtype)
                    loss = self.svi.step(x, y, re)
                    batch_loss[step] = loss
                bl = np.mean(batch_loss)
                _ = sys.stdout.write("\rEpoch: {}/{} loss: {} ".format(epoch+1,epochs, np.round(bl, 3).astype(str)))
                sys.stdout.flush()
            self.posterior_samples = pyro.infer.Predictive(self.model, guide=self.guide, num_samples=num_samples)(torch.tensor(X, dtype=torch.float32), 
                                                                                                                  torch.tensor(Y, dtype=torch.float32),
                                                                                                                  torch.tensor(np.asarray(RE).reshape([-1,1]), dtype=torch.long))

        
        self.scale = self.posterior_samples["scale"].data.squeeze().mean(dim=0).cpu()
        self.mu = self.posterior_samples["mu"].data.squeeze().mean(dim=0).cpu()
        
        if RE is None:
            self.Re = None
        else:
            self.Re = self.posterior_samples["re"].data.squeeze().mean(dim=0).cpu()
    
    def predict(self, newdata, SP=None, RE=None, batch_size = 25, parallel=0, num_samples=10 ):
        dataLoader = self._get_DataLoader(X = newdata, Y = None, SP=SP,RE=RE, batch_size = batch_size, shuffle = False, parallel = parallel, drop_last = False)
        pred = []
        if RE is None:                          
            for step, (x) in enumerate(dataLoader):
                x = x[0].to(self.device, non_blocking=True)
                pred.append(pyro.infer.Predictive(self.model, guide=self.guide, num_samples=num_samples)(XX = x,YY= None)["obs"].squeeze())
        else:
            for step, (x, re) in enumerate(dataLoader):
                x = x.to(self.device, non_blocking=True)
                re = re.to(self.device, non_blocking=True)
                pred.append(pyro.infer.Predictive(self.model, guide=self.guide, num_samples=num_samples)(XX = x,YY= None, RE=re)["obs"].squeeze())  
        return torch.cat(pred, dim=1).data.cpu().numpy()
    
    def logLik(self, X, Y, RE=None, SP=None, batch_size = 25, parallel=0):
        if self.device.type == 'cuda':
            device = self.device.type+ ":" + str(self.device.index)
        else:
            device = 'cpu'

        ww = torch.tensor(self.weights)
        logLik = 0
        dataLoader = self._get_DataLoader(X = X, Y = Y, SP=SP,RE=RE, batch_size = batch_size, shuffle = False, parallel = parallel, drop_last = False)
        pred = []
        if RE is None:                          
            for step, (x, y) in enumerate(dataLoader):
                x = x.to(self.device, non_blocking=True)
                y = y.to(self.device, non_blocking=True)
                mu = torch.nn.functional.linear(x, ww.t())
                logLik += MultivariateProbit(mu, self.scale).log_prob(y).sum().data.cpu().numpy()
                
        else:
            for step, (x, y, re) in enumerate(dataLoader):
                x = x.to(self.device, non_blocking=True)
                re = re.to(self.device, non_blocking=True)
                spatial_re = self.re.gather(0, re.to(self.device, non_blocking=True))
                y = y.to(self.device, non_blocking=True)
                mu = torch.nn.functional.linear(x, ww.t()) + spatial_re
                logLik += MultivariateProbit(mu, self.scale).log_prob(y).sum().data.cpu().numpy()
        return logLik  

    @property
    def covariance(self):
        return self.scale.matmul(self.scale.t()).data.cpu().numpy()

    @property
    def weights(self):
        return self.mu.data.cpu().numpy()

    @property
    def re(self):
        if self.Re is None:
            return None
        else:
            return self.Re.data.cpu().numpy()



import torch
import torch_optimizer as optim
import madgrad

def optimizer_adamax(lr = 0.002,betas = [0.9, 0.999], eps = 1e-08 , weight_decay = 0.0):
    """
    Adamax Optimizer

    # Arguments
    :param lr: float of 1 > 0.0, learning rate
    :param betas: tuple of floats
    :param eps: float of 1 > 0.0
    :param weight_decay: float of 1 > 0.0

    """
    return lambda params: torch.optim.Adamax(params = params, lr = lr, betas = betas, eps = eps, weight_decay = weight_decay)

def optimizer_RMSprop(lr=1e-2, alpha=0.99, eps=1e-8, weight_decay=0, momentum=0, centered=False):
    """
    RMSprop Optimizer

    # Arguments
    :param lr: float of 1 > 0.0, learning rate
    :param alpha: float of 1 > 0.0
    :param eps: float of 1 > 0.0
    :param weight_decay: float of 1 > 0.0
    :param momentum: float of 1 > 0.0
    :param centered: logical of 1

    """
    return lambda params: torch.optim.RMSprop(params = params, lr = lr, alpha=alpha, eps = eps, weight_decay = weight_decay, momentum=momentum, centered=centered)

def optimizer_SGD(lr=1e-2, momentum=0, dampening=0, weight_decay=0, nesterov=False):
    """
    SGD Optimizer

    # Arguments
    :param lr: float of 1 > 0.0, learning rate
    :param momentum: float of 1 > 0.0
    :param dampening: float of 1 > 0.0
    :param eps: float of 1 > 0.0
    :param weight_decay: float of 1 > 0.0
    :param momentum: float of 1 > 0.0
    :param centered: logical of 1

    """
    return lambda params: torch.optim.SGD(params = params, lr = lr, momentum=momentum, dampening=dampening, weight_decay=weight_decay, nesterov=nesterov)


def optimizer_AccSGD(lr=1e-3,
    kappa=1000.0,
    xi=10.0,
    small_const=0.7,
    weight_decay=0):
    return lambda params: optim.AccSGD(params=params,lr=lr,kappa=kappa,xi=xi,small_const=small_const,weight_decay=weight_decay)

def optimizer_AdaBound(lr= 1e-3,
    betas= (0.9, 0.999),
    final_lr = 0.1,
    gamma=1e-3,
    eps= 1e-8,
    weight_decay=0,
    amsbound=True):
    return lambda params: optim.AdaBound(params=params,lr=lr, betas=betas,final_lr=final_lr,gamma=gamma,eps= eps,weight_decay=weight_decay,amsbound=amsbound)

def optimizer_DiffGrad(lr= 1e-3,
    betas=(0.9, 0.999),
    eps=1e-8,
    weight_decay=0):
    return lambda params: optim.DiffGrad(params=params,lr=lr, betas=betas,eps=eps,weight_decay=weight_decay)

def optimizer_madgrad(lr=1e-2, momentum=0, weight_decay=0, eps=1e-6):
    return lambda params: madgrad.MADGRAD(params = params, lr = lr, momentum=momentum, weight_decay=weight_decay, eps=eps)



from .model_sjSDM import Model_sjSDM
from .model_svi import Model_SVI
from .model_LVM import Model_LVM
from .optimizer import *
from .dist_mvp import MVP_logLik
from .utils_fa import covariance, importance


import pytest
from .. import sjSDM_py as fa
import numpy as np

@pytest.fixture
def data():
    def _get(a=5, b=5, c=100):
        return (np.random.randn(c,a), np.random.binomial(1,0.5,[c, b]))
    return _get


@pytest.fixture
def model_base():
    def _get(X, Y,df=2, guide="Delta", 
            scale_mu = 5.0, scale_lf = 1.0,scale_lv=1.0, 
            lr = [0.1], 
            epochs = 2,
            family = "binomial",
            link = "logit",
            batch_size = 25,
            num_samples=100):
        model = fa.Model_LVM()
        model.build(X.shape, Y.shape[1], df=df,guide=guide,
                  scale_mu=scale_mu,scale_lf=scale_lf,scale_lv=scale_lv, family=family,link=link)
        model.fit(X=X, Y=Y,epochs=epochs,batch_size=batch_size, num_samples=num_samples)
        return model
    return _get


@pytest.mark.parametrize("e", [2, 3, 5])
@pytest.mark.parametrize("sp", [3, 6, 11])
@pytest.mark.parametrize("n", [35, 50, 77])
@pytest.mark.parametrize("df", [2,3,5])
@pytest.mark.parametrize("bs", [2, 22])
def test_base(data, model_base, e, sp, n, df, bs):
    X, Y = data(e, sp, n)
    model = model_base(X, Y, df=df, batch_size=bs)
    model.getLogLik(X, Y, batch_size=bs)
    pred = model.predict(X, batch_size=bs)
    assert pred.shape[0] == X.shape[0]


@pytest.mark.parametrize("fam,link", [("binomial", "logit"), ("binomial", "probit"), ("poisson", "log"), ("poisson", "linear")])
def test_fams(data, model_base, fam, link):
    X, Y = data()
    model = model_base(X, Y, family=fam, link=link)
    model.getLogLik(X, Y)
    pred = model.predict(X)
    assert pred.shape[0] == X.shape[0]

@pytest.mark.parametrize("guide", ["LaplaceApproximation", "Delta", "LowRankMultivariateNormal", "DiagonalNormal"])
@pytest.mark.parametrize("lr", [ [0.1], [0.1, 0.1, 0.1] ])
@pytest.mark.parametrize("scales", [(3.0, 1.0, 1.0), (5.0, 3.0, 0.001)])
def test_guide_prior_lrs(data, model_base, guide, lr, scales):
    X, Y = data()
    model = model_base(X, Y, guide=guide, lr=lr, scale_mu=scales[0], scale_lf=scales[1], scale_lv=scales[2])
    model.getLogLik(X, Y)
    pred = model.predict(X)
    assert pred.shape[0] == X.shape[0]   


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sjSDM_py", 
    version="0.1.1",
    author="Maximilian Pichler",
    author_email="Maximilian.Pichler@ur.de",
    description="jSDM package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TheoreticalEcology/s-jSDM",
    packages=setuptools.find_packages(),
    install_requires = [
        "numpy"
    ],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ]
)



import torch
import pyro
from torch.distributions import constraints
from torch.distributions.utils import lazy_property
from pyro.distributions.torch_distribution import TorchDistribution
from pyro.distributions.util import broadcast_shape

class MultivariateProbit(TorchDistribution):
    arg_constraints = {'loc': constraints.real_vector}
    support = constraints.real
    has_rsample = True

    def __init__(self, loc, scale=None, link="probit",sampling=5000, validate_args=None):
        if loc.dim() < 1:
            raise ValueError("loc must be at least one-dimensional.")

        loc_ = loc.unsqueeze(-1)  # temporarily add dim on right

        if scale.dim() < 2:
            raise ValueError("scale must be at least two-dimensional, "
                             "with optional leading batch dimensions")
        self.scale, loc_ = torch.broadcast_tensors(scale, loc_)
        self.scale = scale
        self._covariance_matrix = scale
        self.loc = loc_[..., 0]  # drop rightmost dim
        self.df = scale.shape[-1]
        self.sampling = sampling
        self.alpha = 1.0
        
        if link == "logit":
            self.link = lambda value: torch.sigmoid(value)
            self.alpha = 1.7012
        elif link == "probit":
            self.link = lambda value: torch.distributions.Normal(0.0, 1.0).cdf(value)
        elif link == "linear":
            self.link = lambda value: torch.clamp(value, 0.0, 1.0)

        batch_shape, event_shape = self.loc.shape[:-1], self.loc.shape[-1:]
        super(MultivariateProbit, self).__init__(batch_shape, event_shape, validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(MultivariateProbit, _instance)
        batch_shape = torch.Size(batch_shape)
        loc_shape = batch_shape + self.event_shape
        cov_shape = batch_shape + self.event_shape + torch.Size([self.df])
        new.loc = self.loc.expand(loc_shape)
        if 'scale' in self.__dict__:
            new.scale = self.scale #.expand(cov_shape)
            new.df = self.df
            new._covariance_matrix = self.scale
            new.link = self.link
            new.sampling = self.sampling
            new.alpha = self.alpha
        super(MultivariateProbit, new).__init__(batch_shape,
                                                self.event_shape,
                                                validate_args=False)
        new._validate_args = self._validate_args
        return new

    @lazy_property
    def covariance_matrix(self):
        return (torch.matmul(self._covariance_matrix,
                             self._covariance_matrix.transpose(-1, -2))
                .expand(self._batch_shape + self._event_shape + self._event_shape))

    @property
    def mean(self):
        return self.loc

    def rsample(self, sample_shape=torch.Size()):
        # shape = self._extended_shape(sample_shape)
        # eps = _standard_normal(shape, dtype=self.loc.dtype, device=self.loc.device)
        # return self.loc + _batch_mv(self._unbroadcasted_scale_tril, eps)
        if sample_shape == torch.Size():
            sample_shape = [1]
        shape = self._extended_shape(sample_shape)
        shape2 = torch.tensor(shape[:-1]).prod().view([1])
        
        eps = torch.tensor(0.00001)
        one = torch.tensor(1.0)
        alpha = torch.tensor(self.alpha)
        half = torch.tensor(0.5)
        noise = torch.randn(size = torch.Size([100, shape2, self.df]))
        E = self.link(torch.tensordot(noise, self.scale.t(), dims = 1).add(self.loc).mul(alpha)).mul(one.sub(eps)).add(eps.mul(half)).mean(dim = 0)
        return E.view(shape)
    
    def log_prob(self, value):
        #if self._validate_args:
        #    self._validate_sample(value)
        
        #print(self.scale)
        eps = torch.tensor(0.00001, device=value.device)
        zero = torch.tensor(0.0, device=value.device)
        one = torch.tensor(1.0, device=value.device)
        alpha = torch.tensor(self.alpha, device=value.device)
        half = torch.tensor(0.5, device=value.device)
        
        shape = value.shape
        shape2 = torch.tensor(shape[:-1]).prod().view([1])
        value = value.view(torch.Size([shape2, shape[-1]]))
        
        #print(self.scale.shape)
        
        noise = torch.randn(size = [self.sampling, value.shape[0], self.df], device=value.device)
        E = self.link(torch.tensordot(noise, self.scale.t(), dims = 1).add(self.loc).mul(alpha)).mul(one.sub(eps)).add(eps.mul(half))
        logprob = E.log().mul(value).add(one.sub(E).log().mul(one.sub(value))).neg().sum(dim = 2).neg()
        maxlogprob = logprob.max(dim = 0).values
        Eprob = logprob.sub(maxlogprob).exp().mean(dim = 0)
        loss = Eprob.log().neg().sub(maxlogprob).neg()
        return loss.view(shape[:-1])


def MVP_logLik(Y, pred, sigma, device, batch_size=25, sampling=1000, link="probit", individual=False):

    torch.set_default_tensor_type('torch.FloatTensor')

    if device.type == 'cuda' and torch.cuda.is_available():
        torch.set_default_tensor_type('torch.cuda.FloatTensor')

    if device.type == 'cuda':
        torch.cuda.set_device(device)
        pin_memory = False
        device = device.type+ ":" + str(device.index)
    else:
        pin_memory = True

    if link=="probit":
        link_func = lambda value: torch.distributions.Normal(0.0, 1.0).cdf(value)
    elif link=="linear":
        link_func = lambda value: torch.clamp(value, 0.0, 1.0)
    elif link=="logit":
        link_func = lambda value: torch.sigmoid(value.mul(1.70169))

    data = torch.utils.data.TensorDataset(torch.tensor(Y, dtype=torch.float32, device=torch.device('cpu')), torch.tensor(pred, dtype=torch.float32, device=torch.device('cpu')))
    DataLoader = torch.utils.data.DataLoader(data, batch_size=batch_size, shuffle=False, num_workers=0, pin_memory=pin_memory, drop_last=False)
    torch.cuda.empty_cache()
    sigma=torch.tensor(sigma, dtype=torch.float32, device=torch.device(device))
    logLik = []
    for step, (y, pred) in enumerate(DataLoader):
        y = y.to(device, non_blocking=True)
        pred = pred.to(device, non_blocking=True)

        noise = torch.randn(size = [sampling, batch_size, sigma.shape[1]], device=torch.device(device))
        E = link_func(torch.tensordot(noise, sigma.t(), dims = 1).add(pred)).mul(0.999999).add(0.0000005)
        logprob = E.log().mul(y).add((1.0 - E).log().mul(1.0 - y)).neg().sum(dim = 2).neg()
        maxlogprob = logprob.max(dim = 0).values
        Eprob = logprob.sub(maxlogprob).exp().mean(dim = 0)
        loss = Eprob.log().neg().sub(maxlogprob)
        logLik.append(loss.data)

    if individual is not True:
        logLik = torch.cat(logLik).sum().data.cpu().numpy()
    else:
        logLik = torch.cat(logLik).data.cpu().numpy()
    
    return logLik


import torch
import numpy as np
from typing import Optional, List


@staticmethod
def _device_and_dtype(device, dtype):
    """define device and dtype

    parse device and dtype into torch.dtype and torch.device

    # Arguments
        device: str or torch.device, "cpu" or "gpu"
        dtype: str or torch.dtype, "float32" and "float64" are supported

    # Example

    ```python
        device, dtype = _device_and_dtype(torch.device("cpu"), "float32")
        device, dtype = _device_and_dtype("gpu", "float32")
    ````

    """
    if type(device) is int:
        device = torch.device('cuda:'+str(device))
    
    if type(device) is str:
        device = torch.device(device)

    if type(dtype) is not str:
        return device, dtype
    else:
        if dtype == "float32":
            return device, torch.float32
        if dtype == "float64":
            return device, torch.float64
    return device, torch.float32

def covariance(X):
    # taken from https://github.com/pytorch/pytorch/issues/19037#issuecomment-739002393 
    D = X.shape[-1]
    mean = torch.mean(X, dim=-1).unsqueeze(-1)
    X = X - mean
    return 1/(D-1) * X @ X.transpose(-1, -2)  

def importance(beta: np.ndarray, 
               covX: np.ndarray, 
               sigma: np.ndarray, 
               precision: torch.dtype = torch.float32, 
               covSP: Optional[np.ndarray] = None, 
               betaSP: Optional[np.ndarray] = None) -> dict:
    """Variation partitioning

    Args:
        beta (np.ndarray): environmental coefficients
        covX (np.ndarray): covariance matrix of predictors
        sigma (np.ndarray): square-root matrix of species associations
        precision (torch.dtype, optional): Float or doubles. Defaults to torch.float32.
        covSP (Optional[np.ndarray], optional): covariance matrix of spatial predictors. Defaults to None.
        betaSP (Optional[np.ndarray], optional): spatial coefficients. Defaults to None.

    Returns:
        dict: importances for each functional group
    """               
    beta = torch.tensor(beta.copy(), dtype=precision, device=torch.device("cpu"))
    sigma = torch.tensor(sigma.copy(), dtype=precision, device=torch.device("cpu"))
    sigmaBeta = torch.tensor(covX.copy(), dtype=precision, device=torch.device("cpu"))
    association = sigma.matmul(sigma.t()).add( torch.diag(torch.ones(sigma.shape[0],dtype=precision, device=torch.device("cpu") )) )

    betaCorrected = sigmaBeta.t().matmul(beta)
    Xtotal = torch.einsum("ej, ej -> j", beta, betaCorrected)
    Xsplit = torch.einsum("ej, ej -> je", beta, betaCorrected)

    PredRandom = (association.sum(dim=0) - association.diag()).abs() / association.shape[0]

    if type(betaSP) is np.ndarray:
        betaSP = torch.tensor(betaSP.copy(), dtype=precision, device=torch.device("cpu"))
        sigmaBetaSP = torch.tensor(covSP.copy(), dtype=precision, device=torch.device("cpu"))
        betaSPCorrected = sigmaBetaSP.t().matmul(betaSP)
        SPtotal = torch.einsum("ej, ej -> j", betaSP, betaSPCorrected)
        SPsplit = torch.einsum("ej, ej -> je", betaSP, betaSPCorrected)

        variTotal = Xtotal + PredRandom + SPtotal
        variPartX = Xtotal/variTotal
        variPartSP = SPtotal/variTotal
        variPartRandom = PredRandom/variTotal
        nGroups = Xsplit.shape[1]
        nSP = SPsplit.shape[1]
        variPartXSplit = Xsplit/torch.repeat_interleave(Xsplit.sum(1), nGroups).reshape((Xsplit.shape[0], -1))
        variPartSPSplit = SPsplit/torch.repeat_interleave(SPsplit.sum(1), nSP).reshape((SPsplit.shape[0], -1))
        res = {
            "env" : (variPartX.repeat_interleave(nGroups).reshape([-1, nGroups]) * variPartXSplit).data.cpu().numpy(),
            "spatial" : (variPartSP.repeat_interleave(nSP).reshape([-1, nSP]) * variPartSPSplit).data.cpu().numpy(),
            "biotic" : (variPartRandom).data.cpu().numpy()
        }
    else:
        variTotal = Xtotal + PredRandom
        variPartX = Xtotal/variTotal
        variPartRandom = PredRandom/variTotal
        nGroups = Xsplit.shape[1]
        variPartXSplit = Xsplit/torch.repeat_interleave(Xsplit.sum(1), nGroups).reshape((Xsplit.shape[0], -1))
        res = {
            "env" : (variPartX.repeat_interleave(nGroups).reshape([-1, nGroups]) * variPartXSplit).data.cpu().numpy(),
            "biotic" : (variPartRandom).data.cpu().numpy()
        }
    return res






